#include "framecpp/config.h"

#include "general/Profile.hh"

#include "framecpp/Common/IOStream.hh"
#include "framecpp/Common/Description.hh"
#include "framecpp/Common/SearchContainer.hh"

#include "framecpp/Version6/FrAdcData.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"

#include "framecpp/Version6/PTR_STRUCT.hh"

using FrameCPP::Common::Description;
using FrameCPP::Common::FrameSpec;

using std::ceil;
using std::floor;

//-----------------------------------------------------------------------
// Local functions and variables
//-----------------------------------------------------------------------
static const FrameSpec::Info::frame_object_types s_object_id
= FrameSpec::Info::FSI_FR_ADC_DATA;

static INT_2U MAX_REF = 2;

namespace FrameCPP
{
  namespace Version_6
  {
    //-----------------------------------------------------------------------
    // FrAdcData::data_definition_type
    //-----------------------------------------------------------------------
    FrAdcData::data_definition_type::
    data_definition_type( )
    {
    }

    FrAdcData::data_definition_type::
    data_definition_type( const data_definition_type& Source )
      : data( Source.data )
    {
      copy_core( Source );
    }

    bool FrAdcData::data_definition_type::
    operator==( const data_definition_type& RHS ) const
    {
#if LM_INFO
#define LM_INFO_CHECK(a)						\
      if ( ! ( a == RHS.a ) )						\
      {									\
	std::cerr << "DEBUG: Comparison of " << #a << " failed" << std::endl; \
      }

      LM_INFO_CHECK( name );
      LM_INFO_CHECK( comment);
      LM_INFO_CHECK( channelGroup );
      LM_INFO_CHECK( channelNumber );
      LM_INFO_CHECK( nBits );
      LM_INFO_CHECK( bias );
      LM_INFO_CHECK( slope );
      LM_INFO_CHECK( units );
      LM_INFO_CHECK( fShift );
      LM_INFO_CHECK( timeOffset );
      LM_INFO_CHECK( dataValid );
      LM_INFO_CHECK( phase );
      LM_INFO_CHECK( data );
      LM_INFO_CHECK( aux );
#undef LM_INFO_CHECK
#endif /* LM_INFO */

      const bool	retval =
	( ( this == &RHS ) ||	// Short circuit for comparison to self
	  ( ( name == RHS.name ) &&
	    ( comment == RHS.comment ) &&
	    ( channelGroup == RHS.channelGroup ) &&
	    ( channelNumber == RHS.channelNumber ) &&
	    ( nBits == RHS.nBits ) &&
	    ( sampleRate == RHS.sampleRate ) &&
	    ( bias == RHS.bias ) &&
	    ( slope == RHS.slope ) &&
	    ( units == RHS.units ) &&
	    ( fShift == RHS.fShift ) &&
	    ( timeOffset == RHS.timeOffset ) &&
	    ( dataValid == RHS.dataValid ) &&
	    ( phase == RHS.phase ) &&
	    ( data == RHS.data ) &&
	    ( aux == RHS.aux ) ) );

      return retval;

    }

    bool FrAdcData::
    operator==( const Common::FrameSpec::Object& Obj ) const
    {
      return ( *this == Obj );
    }

    void FrAdcData::data_definition_type::
    copy_core( const data_definition_type& Source )
    {
      name = Source.name;
      comment = Source.comment;
      channelGroup = Source.channelGroup;
      channelNumber = Source.channelNumber;
      nBits = Source.nBits;
      sampleRate = Source.sampleRate;
      bias = Source.bias;
      slope = Source.slope;
      units = Source.units;
      fShift = Source.fShift;
      timeOffset = Source.timeOffset;
      dataValid = Source.dataValid;
      phase = Source.phase;
      aux = Source.aux;
    }

    //-----------------------------------------------------------------------
    //
    //-----------------------------------------------------------------------

    const REAL_4 FrAdcData::DEFAULT_BIAS = FR_ADC_DATA_DEFAULT_BIAS;
    const REAL_4 FrAdcData::DEFAULT_SLOPE = FR_ADC_DATA_DEFAULT_SLOPE;
    const REAL_8 FrAdcData::DEFAULT_FSHIFT = FR_ADC_DATA_DEFAULT_FSHIFT;
    const REAL_8 FrAdcData::DEFAULT_TIME_OFFSET = FR_ADC_DATA_DEFAULT_TIME_OFFSET;
    const INT_2U FrAdcData::DEFAULT_DATA_VALID = FR_ADC_DATA_DEFAULT_DATA_VALID;
    const REAL_4 FrAdcData::DEFAULT_PHASE = FR_ADC_DATA_DEFAULT_PHASE;

    FrAdcData::
    FrAdcData( )
      : FrameSpec::Object( s_object_id, StructDescription( ) )
    {
    }

    FrAdcData::
    FrAdcData( const FrAdcData& Source )
      : FrameSpec::Object( s_object_id, StructDescription( ) ),
	Common::TOCInfo( Source ),
	m_data( Source.m_data )
    {
    }

    FrAdcData::
    FrAdcData( const std::string& name, INT_4U group,
	       INT_4U channel, INT_4U nBits,
	       REAL_8 sampleRate,
	       REAL_4 bias,
	       REAL_4 slope,
	       const std::string& units,
	       REAL_8 fShift,
	       REAL_8 timeOffset,
	       INT_2U dataValid,
	       REAL_4 phase )
      : FrameSpec::Object( s_object_id, StructDescription( ) )
    {
      m_data.name = name;
      m_data.channelGroup = group;
      m_data.channelNumber = channel;
      m_data.nBits = nBits;
      m_data.sampleRate = sampleRate;
      m_data.bias = bias;
      m_data.slope = slope;
      m_data.units = units;
      m_data.fShift = fShift;
      m_data.timeOffset = timeOffset;
      m_data.dataValid = dataValid;
      m_data.phase = phase;
    }

    FrAdcData::
    FrAdcData( Previous::FrAdcData& Source, istream_type* Stream )
      : FrameSpec::Object( s_object_id, StructDescription( ) )
    {
      m_data.name = Source.GetName( );
      m_data.comment = Source.GetComment( );
      m_data.channelGroup = Source.GetChannelGroup( );
      m_data.channelNumber = Source.GetChannelNumber( );
      m_data.nBits = Source.GetNBits( );
      m_data.bias = Source.GetBias( );
      m_data.slope = Source.GetSlope( );
      m_data.units = Source.GetUnits( );
      m_data.sampleRate = Source.GetSampleRate( );
      m_data.timeOffset = REAL_8( Source.GetTimeOffsetS( ) )
	+ REAL_8( Source.GetTimeOffsetN( ) ) * NANOSECOND;
      m_data.fShift = Source.GetFShift( );
      m_data.dataValid = Source.GetDataValid( );

      m_data.phase = FrAdcData::DEFAULT_PHASE;

      if ( Stream )
      {
	//-------------------------------------------------------------------
	// Modify references
	//-------------------------------------------------------------------
	Stream->ReplaceRef( RefData( ), Source.RefData( ), MAX_REF );
	Stream->ReplaceRef( RefAux( ), Source.RefAux( ), MAX_REF );
      }
    }

    FrAdcData::
    FrAdcData( istream_type& Stream )
      : FrameSpec::Object( s_object_id, StructDescription( ) )
    {
      Stream >> m_data.name
	     >> m_data.comment
	     >> m_data.channelGroup
	     >> m_data.channelNumber
	     >> m_data.nBits
	     >> m_data.bias
	     >> m_data.slope
	     >> m_data.units
	     >> m_data.sampleRate
	     >> m_data.timeOffset
	     >> m_data.fShift
	     >> m_data.phase
	     >> m_data.dataValid
	     >> m_data.data
	     >> m_data.aux
	;
      Stream.Next( this );
    }

    FrAdcData::
    ~FrAdcData( )
    {
    }

    const std::string& FrAdcData::
    GetName() const
    {
      return m_data.name;
    }


    FrameCPP::cmn_streamsize_type FrAdcData::
    Bytes( const Common::StreamBase& Stream ) const
    {
      return
	m_data.name.Bytes( ) +
	m_data.comment.Bytes( ) +
	sizeof( m_data.channelGroup ) +
	sizeof( m_data.channelNumber ) +
	sizeof( m_data.nBits ) +
	sizeof( m_data.bias ) +
	sizeof( m_data.slope ) +
	m_data.units.Bytes( ) +
	sizeof( m_data.sampleRate ) +
	sizeof( m_data.timeOffset ) +
	sizeof( m_data.fShift ) +
	sizeof( m_data.phase ) +
	sizeof( m_data.dataValid ) +
	Stream.PtrStructBytes( ) +	// aux
	Stream.PtrStructBytes( ) +	// data
	Stream.PtrStructBytes( )	// next
	;
    }

    FrAdcData* FrAdcData::
    Create( istream_type& Stream ) const
    {
      return new FrAdcData( Stream );
    }

    FrAdcData& FrAdcData::
    Merge( const FrAdcData& RHS )
    {
      //:TODO: Need to implement Merge routine
      std::string msg( "Merge currently not implemented for " );
      msg += StructName( );

      throw std::domain_error( msg );
      return *this;
    }

    const char* FrAdcData::
    ObjectStructName( ) const
    {
      return StructName( );
    }


    const Description* FrAdcData::
    StructDescription( )
    {
      static Description ret;

      if ( ret.size( ) == 0 )
      {
	ret( FrSH( FrAdcData::StructName( ), s_object_id,
		   "AdcData  Data Structure" ) );

	ret( FrSE( "name", "STRING",
		   "" ) );
	ret( FrSE( "comment", "STRING",
		   "" ) );
	ret( FrSE( "channelGroup", "INT_4U",
		   "" ) );
	ret( FrSE( "channelNumber", "INT_4U",
		   "" ) );
	ret( FrSE( "nBits", "INT_4U",
		   "" ) );
	ret( FrSE( "bias", "REAL_4",
		   "" ) );
	ret( FrSE( "slope", "REAL_4",
		   "" ) );
	ret( FrSE( "units", "STRING",
		   "" ) );
	ret( FrSE( "sampleRate", "REAL_8",
		   "" ) );
	ret( FrSE( "timeOffset", "REAL_8",
		   "" ) );
	ret( FrSE( "fShift", "REAL_8",
		   "" ) );
	ret( FrSE( "phase", "REAL_4",
		   "" ) );
	ret( FrSE( "dataValid", "INT_2U",
		   "" ) );

	ret( FrSE( "data", PTR_STRUCT::Desc( FrVect::StructName( ) ),
		   ""  ) );
	ret( FrSE( "aux", PTR_STRUCT::Desc( FrVect::StructName( ) ),
		   ""  ) );

	ret( FrSE( "next",	PTR_STRUCT::Desc( FrAdcData::StructName( ) ),
		   ""  ) );
      }

      return &ret;
    }

    std::auto_ptr< FrAdcData > FrAdcData::
    SubFrAdcData( REAL_8 Offset, REAL_8 Dt ) const
    {
      //---------------------------------------------------------------------
      // Sanity checks
      //---------------------------------------------------------------------
      // Verify the existance of only a single FrVect component
      //---------------------------------------------------------------------
      if ( m_data.data.size( ) != 1 )
      {
	std::ostringstream	msg;

	msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	    << "data vectors of length " << m_data.data.size( )
	    << " currently are not supported";
	throw std::logic_error( msg.str( ) );
      }
      INT_4U	num_elements = m_data.data[0]->GetDim( 0 ).GetNx( );
      //---------------------------------------------------------------------
      //:TODO: Ensure the dimension of the FrVect is 1
      //---------------------------------------------------------------------
      //---------------------------------------------------------------------
      //:TODO: Ensure the requested range is valid
      //---------------------------------------------------------------------
      if ( Offset > ( m_data.timeOffset +
		      ( num_elements / m_data.sampleRate ) ) )
      {
	std::ostringstream msg;
	msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	    << "The value for the parameter Offset (" << Offset
	    << ") is out of range ([0-"
	    << ( m_data.timeOffset + ( num_elements / m_data.sampleRate ) )
	    << "))";
	throw std::range_error( msg.str( ) );
      }
      //---------------------------------------------------------------------
      // Start moving of data
      //---------------------------------------------------------------------
      std::auto_ptr< FrAdcData >	retval( new FrAdcData );

      // Copy in the header information
      retval->m_data.copy_core( this->m_data );

      //:TODO: Get the slice of FrVect
      REAL_8	startSampleReal =
	( Offset * retval->m_data.sampleRate ) -
	( retval->m_data.timeOffset * retval->m_data.sampleRate );

      INT_4U	endSample =
	INT_4U( ceil ( startSampleReal +
		       ( retval->m_data.sampleRate * Dt ) ) );
      INT_4U	startSample = ( startSampleReal > REAL_8( 0 ) )
	? INT_4U( floor( startSampleReal ) )
	: INT_4U( 0 );

      {
	General::SharedPtr< FrVect >
	  subvect( this->m_data.data[ 0 ]->
		   SubFrVect( startSample, endSample ).release( ) );

	retval->m_data.data.append( subvect );
      }
      //---------------------------------------------------------------------
      // Shift the offset
      //---------------------------------------------------------------------
      if ( startSampleReal > 0 )
      {
	// There is no gap
	retval->m_data.timeOffset = REAL_8( 0 );
      }
      else
      {
	retval->m_data.timeOffset += Offset;
      }
      //---------------------------------------------------------------------
      // Return the new FrAdcData
      //---------------------------------------------------------------------
      return retval;
    }
   
    void FrAdcData::
    Write( ostream_type& Stream ) const
    {
      try
      {
	Stream << m_data.name
	       << m_data.comment
	       << m_data.channelGroup
	       << m_data.channelNumber
	       << m_data.nBits
	       << m_data.bias
	       << m_data.slope
	       << m_data.units
	       << m_data.sampleRate
	       << m_data.timeOffset
	       << m_data.fShift
	       << m_data.phase
	       << m_data.dataValid
	       << m_data.data
	       << m_data.aux
	  ;
	WriteNext( Stream );
      }
      catch ( const std::exception& Exception )
      {
	std::cerr << "DEBUG: Error Writing FrAdcData: " << m_data.name
		  << " exception: " << Exception.what( )
		  << std::endl;
	throw; // rethrow
      }
    }

    void FrAdcData::
    TOCQuery( int InfoClass, ... ) const
    {
      using Common::TOCInfo;

      va_list	vl;
      va_start( vl, InfoClass );
      while ( InfoClass != TOCInfo::IC_EOQ )
      {
	int data_type = va_arg( vl, int );
	switch( data_type )
	{
	case TOCInfo::DT_STRING_2:
	  {
	    STRING* data = va_arg( vl, STRING* );
	    switch( InfoClass )
	    {
	    case TOCInfo::IC_NAME:
	      *data = GetName( );
	      break;
	    default:
	      goto cleanup;
	      break;
	    }
	  }
	  break;
	case TOCInfo::DT_INT_2U:
	  {
	    INT_2U* data = va_arg( vl, INT_2U* );
	    switch( InfoClass )
	    {
	    case TOCInfo::IC_DATA_VALID:
	      *data = GetDataValid( );
	      break;
	    default:
	      goto cleanup;
	      break;
	    }
	  }
	  break;
	case TOCInfo::DT_INT_4U:
	  {
	    INT_4U* data = va_arg( vl, INT_4U* );
	    switch( InfoClass )
	    {
	    case TOCInfo::IC_CHANNEL_ID:
	      *data = GetChannelNumber( );
	      break;
	    case TOCInfo::IC_GROUP_ID:
	      *data = GetChannelGroup( );
	      break;
	    default:
	      goto cleanup;
	      break;
	    }
	  }
	  break;
	default:
	  // Stop processing
	  goto cleanup;
	}
	InfoClass = va_arg( vl, int );
      }
    cleanup:
      va_end( vl );
    }

    FrAdcData::demote_ret_type FrAdcData::
    demote( INT_2U Target,
	    demote_arg_type Obj,
	    istream_type* Stream ) const
    {
      if ( Target >= DATA_FORMAT_VERSION )
      {
	return Obj;
      }
      try
      {
	//-------------------------------------------------------------------
	// Copy non-reference information
	//-------------------------------------------------------------------

	INT_4S	Seconds = INT_4S( GetTimeOffset( ) );
	INT_4U	Nanoseconds = INT_4U( ( GetTimeOffset( ) - Seconds ) * NANOSECOND_MULTIPLIER );

	// Do actual down conversion
	General::SharedPtr< Previous::FrAdcData >
	  retval( new Previous::FrAdcData( GetName( ), GetChannelGroup( ),
					   GetChannelNumber( ), GetNBits( ),
					   GetSampleRate( ),
					   GetBias( ),
					   GetSlope( ),
					   GetUnits( ),
					   GetFShift( ),
					   Seconds, Nanoseconds,
					   GetDataValid( )
					   ) )
	  ;
	reinterpret_cast< Previous::FrAdcData* >(  retval.get( ) )->AppendComment( GetComment( ) );
	if ( Stream )
	{
	  //-----------------------------------------------------------------
	  // Modify references
	  //-----------------------------------------------------------------
	  Stream->ReplaceRef( retval->RefData( ), RefData( ), MAX_REF );
	  Stream->ReplaceRef( retval->RefAux( ), RefAux( ), MAX_REF );
	}
	//-------------------------------------------------------------------
	// Return demoted object
	//-------------------------------------------------------------------
	return retval;
      }
      catch( ... )
      {
      }
      throw
	Unimplemented( "Object* FrAdcData::demote( Object* Obj ) const",
		       DATA_FORMAT_VERSION, __FILE__, __LINE__ );
    }

    FrAdcData::promote_ret_type FrAdcData::
    promote( INT_2U Target,
	     promote_arg_type Obj,
	     istream_type* Stream ) const
    {
      return Promote( Target, Obj, Stream );
    }
  } // namespace - Version_6
} // namespace - FrameCPP
